from datetime import datetime
from sqlalchemy import Column, Integer, String, Text, DateTime, ForeignKey, func, or_, and_, desc, case
from sqlalchemy.orm import relationship, joinedload
from models.base import Base, Session, logger
from models.media_download import MediaDownload
from models.chat import Chat
from models.author import Author

class Message(Base):
    __tablename__ = 'messages'

    id = Column(Integer, primary_key=True)
    body = Column(Text, nullable=True)
    caption = Column(Text, nullable=True)
    message_type = Column(String(20))
    timestamp = Column(DateTime, index=True)
    author_id = Column(Integer, ForeignKey('authors.id'))
    number = Column(String(50), nullable=True)
    chat_id = Column(String(50), ForeignKey('chats.chat_id'))
    duplicate_count = Column(Integer, default=0)
    local_url = Column(String(500), nullable=True)
    file_size = Column(Integer, nullable=True)
    file_hash = Column(String(100), nullable=True, index=True)
    created_at = Column(DateTime, default=datetime.utcnow, nullable=False)
    last_updated = Column(DateTime, default=datetime.utcnow, onupdate=datetime.utcnow)

    author = relationship("Author", back_populates="messages")
    chat = relationship("Chat",
                       back_populates="messages",
                       foreign_keys=[chat_id])  # Explicitly specify the foreign key
    media_downloads = relationship("MediaDownload", back_populates="message")

    def to_dict(self):
        return {
            'id': self.id,
            'body': self.body,
            'caption': self.caption,
            'messageType': self.message_type,
            'timestamp': self.timestamp.isoformat() if self.timestamp else None,
            'authorId': self.author_id,
            'author': self.author.name if self.author else None,
            'chatId': self.chat_id,
            'localUrl': self.local_url,
            'fileSize': self.file_size,
            'fileHash': self.file_hash,
            'duplicateCount': self.duplicate_count or 0,
            'created_at': self.created_at.isoformat() if self.created_at else None,
            'last_updated': self.last_updated.isoformat() if self.last_updated else None
        }

    @classmethod
    def _get_session_with_error_handling(cls, operation_name):
        session = Session()
        try:
            return session
        except Exception as e:
            logger.error(f"Error in {operation_name}: {str(e)}")
            session.close()
            raise

    @classmethod
    def save_messages(cls, messages_df):
        saved_count = duplicate_count = 0
        session = cls._get_session_with_error_handling("save_messages")

        try:
            for _, row in messages_df.iterrows():
                if duplicate := cls.find_duplicate_message(
                    body=row.get('body'),
                    caption=row.get('caption'),
                    message_type=row['message_type'],
                    file_hash=row.get('file_hash')
                ):
                    cls.update_duplicate_count(duplicate, row['timestamp'])
                    duplicate_count += 1
                    continue

                # Create a dictionary of message attributes from the row
                message_data = {k: v for k, v in row.items() if hasattr(cls, k)}

                # Explicitly set created_at to current time
                current_time = datetime.utcnow()
                message_data['created_at'] = current_time

                # Create the message object and add it to the session
                message = cls(**message_data)
                session.add(message)

                logger.info(f"Created new message with created_at: {current_time}")

                if row.get('local_url'):
                    session.add(MediaDownload(
                        media_type=row['message_type'].upper(),
                        download_status='COMPLETED',
                        local_url=row['local_url'],
                        message=message
                    ))

                if chat := session.query(Chat).filter_by(chat_id=row['chat_id']).first():
                    chat.message_count += 1
                    chat.last_message = message.id
                    chat.last_updated = datetime.utcnow()

                saved_count += 1

            session.commit()
            return saved_count, duplicate_count
        finally:
            session.close()

    @classmethod
    def get_messages(cls, page, per_page):
        session = cls._get_session_with_error_handling("get_messages")
        try:
            return (
                [msg.to_dict() for msg in session.query(cls)
                    .options(joinedload(cls.author))
                    .order_by(cls.timestamp.desc())
                    .offset((page - 1) * per_page)
                    .limit(per_page)],
                session.query(cls).count(),
                session.query(cls.chat_id.distinct()).count()
            )
        finally:
            session.close()

    @classmethod
    def create_message(cls, data):
        """Create a new message"""
        session = cls._get_session_with_error_handling("create_message")
        try:
            # Ensure created_at is set to current time
            current_time = datetime.utcnow()
            if 'created_at' not in data:
                data['created_at'] = current_time

            # Create the message object
            message = cls(**data)
            session.add(message)

            # Update chat message count and last_updated if chat exists
            if 'chat_id' in data and data['chat_id']:
                if chat := session.query(Chat).filter_by(chat_id=data['chat_id']).first():
                    chat.message_count += 1
                    chat.last_message = message.id
                    chat.last_updated = current_time

            session.commit()
            logger.info(f"Created new message with ID: {message.id}, created_at: {current_time}")
            return message
        except Exception as e:
            session.rollback()
            logger.error(f"Error creating message: {str(e)}")
            raise
        finally:
            session.close()

    @classmethod
    def update_message(cls, message_id, data):
        """Update an existing message"""
        session = cls._get_session_with_error_handling("update_message")
        try:
            # Find the message
            message = session.query(cls).get(message_id)
            if not message:
                logger.warning(f"Message not found with ID: {message_id}")
                return None

            # Update the message fields
            updated = False
            updateable_fields = ['body', 'caption', 'message_type']

            for field in updateable_fields:
                if field in data:
                    old_value = getattr(message, field)
                    new_value = data[field]
                    if old_value != new_value:
                        setattr(message, field, new_value)
                        updated = True
                        logger.info(f"Updated message {message_id} field {field} from '{old_value}' to '{new_value}'")

            # Explicitly update the last_updated timestamp if any fields were changed
            if updated:
                current_time = datetime.utcnow()
                message.last_updated = current_time
                logger.info(f"Updated message {message_id} with new last_updated: {current_time}")

            session.commit()
            return message
        except Exception as e:
            session.rollback()
            logger.error(f"Error updating message: {str(e)}")
            raise
        finally:
            session.close()

    @classmethod
    def search_messages(cls, query, search_type, page, per_page):
        session = Session()
        try:
            offset = (page - 1) * per_page
            base_query = cls.query.join(Author)  # Join with Author table

            if query:
                query_lower = query.lower()
                if search_type == 'all':
                    # Using case statements to prioritize different match types
                    base_query = base_query.add_columns(
                        case(
                            # Caption exact match (highest priority)
                            [(cls.caption.ilike(query_lower), 300)],
                            # Caption partial match
                            [(cls.caption.ilike(f'%{query_lower}%'), 200)],
                            # Author name match
                            [(Author.name.ilike(f'%{query_lower}%'), 100)],
                            # Body text match (lowest priority)
                            [(cls.body.ilike(f'%{query_lower}%'), 50)],
                            else_=0
                        ).label('relevance')
                    ).filter(
                        or_(
                            cls.caption.ilike(f'%{query_lower}%'),
                            Author.name.ilike(f'%{query_lower}%'),
                            cls.body.ilike(f'%{query_lower}%')
                        )
                    ).order_by(desc('relevance'), cls.timestamp.desc())
                elif search_type == 'name':
                    base_query = base_query.filter(Author.name.ilike(f'%{query}%'))
                elif search_type == 'chat':
                    base_query = base_query.filter(cls.body.ilike(f'%{query}%'))
                elif search_type in ['image', 'video']:
                    base_query = base_query.filter(
                        and_(
                            cls.message_type == search_type,
                            or_(
                                cls.caption.ilike(f'%{query}%'),
                                Author.name.ilike(f'%{query}%')
                            )
                        )
                    )

            total_messages = base_query.count()
            messages = [msg.to_dict() for msg in base_query
                       .offset(offset)
                       .limit(per_page)
                       .all()]

            return messages, total_messages
        finally:
            session.close()

    @classmethod
    def find_duplicate_message(cls, body=None, caption=None, message_type=None, file_hash=None):
        session = cls._get_session_with_error_handling("find_duplicate_message")
        try:
            query = session.query(cls).with_for_update()

            if message_type in ["image", "video"]:
                # Only consider file_hash for media duplicates, not caption
                if file_hash and (duplicate := query.filter(
                    cls.file_hash == file_hash,
                    cls.message_type == message_type,
                    cls.local_url.isnot(None)
                ).first()):
                    logger.info(f"Found duplicate {message_type} by file_hash: {file_hash}")
                    return duplicate

                # Removed caption-based duplicate detection for media files
                # as it's too aggressive and causes too many false positives

            elif message_type == "chat" and body and body.strip():
                # For chat messages, check if we have an exact match in the last 24 hours
                # This reduces false positives for common messages
                from datetime import timedelta
                recent_time = datetime.utcnow() - timedelta(hours=24)

                duplicate = query.filter(
                    cls.message_type == 'chat',
                    func.lower(cls.body) == func.lower(body.strip()),
                    cls.timestamp > recent_time
                ).first()

                if duplicate:
                    logger.info(f"Found duplicate chat message: {body[:50]}...")
                    return duplicate

            return None
        finally:
            session.close()

    @classmethod
    def update_duplicate_count(cls, message, new_timestamp):
        session = cls._get_session_with_error_handling("update_duplicate_count")
        try:
            if message := session.query(cls).with_for_update().get(message.id):
                message.duplicate_count = (message.duplicate_count or 1) + 1
                if isinstance(new_timestamp, str):
                    try:
                        # First try the exact format
                        new_timestamp = datetime.strptime(new_timestamp, '%Y-%m-%dT%H:%M:%S')
                    except ValueError:
                        try:
                            # If that fails, try parsing with milliseconds
                            ts_parts = new_timestamp.split('.')
                            if len(ts_parts) == 2:
                                # Parse the base part without milliseconds
                                new_timestamp = datetime.strptime(ts_parts[0], '%Y-%m-%dT%H:%M:%S')
                            else:
                                # If all parsing fails, use current time
                                new_timestamp = datetime.now()
                        except Exception:
                            # If all parsing fails, use current time
                            new_timestamp = datetime.now()
                if new_timestamp > message.timestamp:
                    message.timestamp = new_timestamp
                    # Also update the last_updated field
                    message.last_updated = datetime.utcnow()
                    logger.info(f"Updated message {message.id} timestamp and last_updated")
                session.commit()
                return True
            return False
        finally:
            session.close()

    @classmethod
    def get_messages_by_author(cls, page=1, per_page=100):
        session = Session()
        try:
            # First, get authors with their latest message timestamp
            latest_messages_subquery = (
                session.query(
                    cls.author_id,
                    func.max(cls.timestamp).label('latest_timestamp')
                )
                .group_by(cls.author_id)
                .subquery()
            )

            # Get total number of authors
            total_authors = session.query(Author.id).distinct().count()

            # Calculate offset
            authors_to_skip = (page - 1) * per_page

            # Get paginated authors ordered by their latest message timestamp
            authors = (session.query(Author, latest_messages_subquery.c.latest_timestamp)
                      .join(latest_messages_subquery,
                            Author.id == latest_messages_subquery.c.author_id)
                      .order_by(latest_messages_subquery.c.latest_timestamp.desc())  # Primary ordering by timestamp
                      .offset(authors_to_skip)
                      .limit(per_page)
                      .all())

            result = []
            for author, latest_timestamp in authors:
                # Get the 10 most recent messages for this author
                recent_messages = (session.query(cls)
                                 .filter(cls.author_id == author.id)
                                 .order_by(cls.timestamp.desc())
                                 .limit(10)
                                 .all())

                # Convert messages to dict format
                messages_data = []
                for msg in recent_messages:
                    message_dict = {
                        'id': msg.id,
                        'body': msg.body,
                        'message_type': msg.message_type,
                        'caption': msg.caption,
                        'local_url': msg.local_url,
                        'file_size': msg.file_size,
                        'file_hash': msg.file_hash,
                        'timestamp': msg.timestamp.isoformat() if msg.timestamp else None
                    }
                    messages_data.append(message_dict)

                author_data = {
                    'author_id': author.id,
                    'name': author.name,
                    'number': author.number,
                    'message_count': len(messages_data),
                    'latest_timestamp': latest_timestamp.isoformat() if latest_timestamp else None,
                    'messages': messages_data
                }
                result.append(author_data)

            return {
                'authors': result,
                'pagination': {
                    'total_authors': total_authors,
                    'has_next': (page * per_page) < total_authors,
                    'current_page': page,
                    'total_pages': (total_authors + per_page - 1) // per_page
                }
            }
        finally:
            session.close()
